#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import math
import numpy as np
import torch

from monai.transforms import Compose
from pathlib import Path
from pl_bolts.models.self_supervised import SimCLR
from timm.models import swin_tiny_patch4_window7_224
from timm.models.swin_transformer import SwinTransformer
from torch import Tensor as T, nn
from torch.utils.checkpoint import checkpoint, checkpoint_sequential
from torchvision.models import resnet18, resnet50, densenet121
from torchvision.models.resnet import ResNet
from torchvision.models.densenet import DenseNet
from typing import Any, Callable, Optional, Sequence, Tuple

from health_cpath.utils.layer_utils import get_imagenet_preprocessing, load_weights_to_model, setup_feature_extractor


class TileEncoder(nn.Module):
    """Base tile encoder class for use in dataset transforms or as part of a bigger model"""

    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        input_dim: Optional[Sequence[int]] = None,
        use_activation_checkpointing: bool = False,
    ) -> None:
        """The `TileEncoder` constructor should be called after setting any attributes needed in
        `_get_preprocessing()` or `_get_encoder()`.

        :param tile_size: Tile width/height, in pixels (default=224).
        :param n_channels: Number of channels in the tile (default=3).
        :param input_dim: Input shape, to override default of `(n_channels, tile_size, tile_size)`.
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations. Note that this is only
            supported for Resnet and Swin Transformer encoders at the moment. The encoder class needs to override
            `custom_forward()` to support checkpointing (default=False).
        """
        super().__init__()
        if input_dim is None:
            if tile_size == 0:
                raise ValueError("Either input_dim or tile_size must be specified")
            input_dim = (n_channels, tile_size, tile_size)
        self.input_dim = tuple(input_dim)

        self.preprocessing_fn = self._get_preprocessing()
        self.feature_extractor_fn, self.num_encoding = self._get_encoder()
        self.use_activation_checkpointing = use_activation_checkpointing

    def _get_preprocessing(self) -> Callable:
        return Compose([])

    def _get_encoder(self) -> Tuple[Callable, int]:
        raise NotImplementedError

    def custom_forward(self, images: torch.Tensor) -> torch.Tensor:
        """A custom forward pass that uses checkpointing to save memory."""
        raise NotImplementedError("Checkpointing is not supported for this encoder.")

    def forward(self, images: torch.Tensor) -> torch.Tensor:
        prep_images = self.preprocessing_fn(images)
        if self.use_activation_checkpointing:
            return self.custom_forward(prep_images)
        return self.feature_extractor_fn(prep_images)


class IdentityEncoder(TileEncoder):
    """Dummy encoder that just flattens the input"""

    def _get_encoder(self) -> Tuple[Callable, int]:
        return nn.Flatten(), np.prod(self.input_dim)  # type: ignore


class ImageNetEncoder(TileEncoder):
    """Feature extractor pretrained for classification on ImageNet"""

    def __init__(
        self,
        feature_extraction_model: Callable[..., nn.Module],
        tile_size: int = 224,
        n_channels: int = 3,
        apply_imagenet_preprocessing: bool = True,
        use_activation_checkpointing: bool = False,
    ) -> None:
        """
        :param feature_extraction_model: A function accepting a `pretrained` keyword argument that
            returns a classifier pretrained on ImageNet, such as the ones from `torchvision.models.*`.
        :param tile_size: Tile width/height, in pixels (default=224).
        :param n_channels: Number of channels in the tile (default=3).
        :param apply_imagenet_preprocessing: Whether to apply ImageNet preprocessing to the input (default=True).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations. Note that this is only
            supported for Resnet and Swin Transformer encoders at the moment. The encoder class needs to override
            `custom_forward()` to support checkpointing (default=False).
        """
        self.create_feature_extractor_fn = feature_extraction_model
        self.apply_imagenet_preprocessing = apply_imagenet_preprocessing
        super().__init__(
            tile_size=tile_size, n_channels=n_channels, use_activation_checkpointing=use_activation_checkpointing
        )

    def _get_preprocessing(self) -> Callable:
        base_preprocessing = super()._get_preprocessing()
        if self.apply_imagenet_preprocessing:
            return Compose([get_imagenet_preprocessing(), base_preprocessing]).flatten()
        return base_preprocessing

    def _get_encoder(self) -> Tuple[torch.nn.Module, int]:
        pretrained_model = self.create_feature_extractor_fn(pretrained=True)
        return setup_feature_extractor(pretrained_model, self.input_dim)


class BaseCheckpointingMixin:
    """Mixin class for checkpointing activations."""

    def __init__(
        self,
        feature_extractor_fn: Any,
        batchnorm_momentum: Optional[float] = None,
        checkpoint_segments_size: int = 2,
        use_activation_checkpointing: bool = False,
    ) -> None:
        """
        :param feature_extractor_fn: A feature extractor model.
        :param batchnorm_momentum: An optional momentum value to use for batch norm layers statistics updates when
            `use_activation_checkpointing` is True. If None (default), sqrt of the default momentum retrieved from the
            model is used to avoid running statistics from going out of sync due to activations checkpointing.
        :param checkpoint_segments_size: If checkpointing, the size of checkpointed segments in sequential layers
            (default=2).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        """
        self.feature_extractor_fn = feature_extractor_fn
        self.checkpoint_segments_size = checkpoint_segments_size
        self.batchnorm_momentum = batchnorm_momentum
        if use_activation_checkpointing:
            self._set_batch_norm_momentum()
        self.validate()

    def validate(self) -> None:
        """Validation checks for the feature extractor model."""
        pass

    def _set_batch_norm_momentum_in_layer(self, layer: nn.Module) -> None:
        """Set the momentum of batch norm layers in the given layer to the value of `self.batchnorm_momentum`.

        :param layer: The layer to set the batch norm momentum for.
        """
        if isinstance(layer, nn.BatchNorm2d):
            assert self.batchnorm_momentum is not None, "batchnorm_momentum must be set"
            layer.momentum = self.batchnorm_momentum

    def _set_all_batch_norm_momentum_in_block(self, layer_block: nn.Sequential) -> None:
        """Set the momentum of batch norm layers in the given block to the value of `self.batchnorm_momentum`.

        :param layer_block: A block of layers to set the batch norm momentum for.
        """
        for sub_layer in layer_block:
            if len(sub_layer._modules) > 0:
                for _, layer in sub_layer._modules.items():
                    self._set_batch_norm_momentum_in_layer(layer)
            else:
                self._set_batch_norm_momentum_in_layer(sub_layer)

    def _set_batch_norm_momentum(self) -> None:
        """Set the momentum of batch norm layers in the feature extractor model"""
        raise NotImplementedError

    def custom_forward(self, images: torch.Tensor) -> torch.Tensor:
        """Custom forward pass that uses activation checkpointing to save memory."""
        raise NotImplementedError


class ResNetCheckpointingMixin(BaseCheckpointingMixin):
    """Mixin class for checkpointing activations in ResNet-based encoders."""

    def validate(self) -> None:
        """Validate that the feature extractor is a ResNet model."""
        assert isinstance(self.feature_extractor_fn, ResNet), "Expected ResNet model for feature_extractor_fn argument."

    def _set_batch_norm_momentum(self) -> None:
        """Set the momentum of batch norm layers in the ResNet model to avoid running statistics from going out of
        sync due to activations checkpointing. The forward pass is applied twice which results in double updates of
        these statistics. We can workaround that by using sqrt of default momentum retrieved from the
        feature_extractor_fn.
        """
        if self.batchnorm_momentum is None:
            self.batchnorm_momentum = math.sqrt(self.feature_extractor_fn.bn1.momentum)

        # Set momentum for the first batch norm layer
        self.feature_extractor_fn.bn1.momentum = self.batchnorm_momentum

        # Fetch all nested batch norm layers and set momentum
        self._set_all_batch_norm_momentum_in_block(self.feature_extractor_fn.layer1)
        self._set_all_batch_norm_momentum_in_block(self.feature_extractor_fn.layer2)
        self._set_all_batch_norm_momentum_in_block(self.feature_extractor_fn.layer3)
        self._set_all_batch_norm_momentum_in_block(self.feature_extractor_fn.layer4)

    def custom_forward(self, images: torch.Tensor) -> torch.Tensor:
        """Custom forward pass that uses activation checkpointing to save memory."""
        segments = self.checkpoint_segments_size
        first_layers = [
            self.feature_extractor_fn.conv1,
            self.feature_extractor_fn.bn1,
            self.feature_extractor_fn.relu,
            self.feature_extractor_fn.maxpool,
        ]
        images = checkpoint_sequential(first_layers, segments, images)
        images = checkpoint_sequential(self.feature_extractor_fn.layer1, segments, images)
        images = checkpoint_sequential(self.feature_extractor_fn.layer2, segments, images)
        images = checkpoint_sequential(self.feature_extractor_fn.layer3, segments, images)
        images = checkpoint_sequential(self.feature_extractor_fn.layer4, segments, images)
        # checkpoint sequantial is used for nn.Sequential layers, we use checkpoint for nn.Module like fc and avgpool
        images = checkpoint(self.feature_extractor_fn.avgpool, images)
        images = torch.flatten(images, 1)
        images = checkpoint(self.feature_extractor_fn.fc, images)
        return images


class Resnet18(ResNetCheckpointingMixin, ImageNetEncoder):
    """ResNet18 encoder with imagenet preprocessing."""

    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        use_activation_checkpointing: bool = False,
        checkpoint_segments_size: int = 2,
        batchnorm_momentum: Optional[float] = None,
    ) -> None:
        """
        :param tile_size: The size of the input tiles (default=224).
        :param n_channels: The number of channels in the input tiles (default=3).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        :param batchnorm_momentum: An optional momentum value to use for batch norm layers statistics updates when
            `use_activation_checkpointing` is True. If None (default), sqrt of the default momentum retrieved from the
            model is used to avoid running statistics from going out of sync due to activations checkpointing.
        """
        ImageNetEncoder.__init__(
            self,
            feature_extraction_model=resnet18,
            tile_size=tile_size,
            n_channels=n_channels,
            apply_imagenet_preprocessing=True,
            use_activation_checkpointing=use_activation_checkpointing,
        )
        ResNetCheckpointingMixin.__init__(
            self, self.feature_extractor_fn, batchnorm_momentum, checkpoint_segments_size, use_activation_checkpointing
        )


class Resnet18_NoPreproc(ResNetCheckpointingMixin, ImageNetEncoder):
    """ResNet18 encoder without imagenet preprocessing."""

    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        use_activation_checkpointing: bool = False,
        checkpoint_segments_size: int = 2,
        batchnorm_momentum: Optional[float] = None,
    ) -> None:
        """
        :param tile_size: The size of the input tiles (default=224).
        :param n_channels: The number of channels in the input tiles (default=3).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        :param batchnorm_momentum: An optional momentum value to use for batch norm layers statistics updates when
            `use_activation_checkpointing` is True. If None (default), sqrt of the default momentum retrieved from the
            model is used to avoid running statistics from going out of sync due to activations checkpointing.
        """
        ImageNetEncoder.__init__(
            self,
            feature_extraction_model=resnet18,
            tile_size=tile_size,
            n_channels=n_channels,
            apply_imagenet_preprocessing=False,
            use_activation_checkpointing=use_activation_checkpointing,
        )
        ResNetCheckpointingMixin.__init__(
            self, self.feature_extractor_fn, batchnorm_momentum, checkpoint_segments_size, use_activation_checkpointing
        )


class Resnet50(ResNetCheckpointingMixin, ImageNetEncoder):
    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        use_activation_checkpointing: bool = False,
        checkpoint_segments_size: int = 2,
        batchnorm_momentum: Optional[float] = None,
    ) -> None:
        """
        :param tile_size: The size of the input tiles (default=224).
        :param n_channels: The number of channels in the input tiles (default=3).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        :param batchnorm_momentum: An optional momentum value to use for batch norm layers statistics updates when
            `use_activation_checkpointing` is True. If None (default), sqrt of the default momentum retrieved from the
            model is used to avoid running statistics from going out of sync due to activations checkpointing.
        """
        ImageNetEncoder.__init__(
            self,
            feature_extraction_model=resnet50,
            tile_size=tile_size,
            n_channels=n_channels,
            apply_imagenet_preprocessing=True,
            use_activation_checkpointing=use_activation_checkpointing,
        )
        ResNetCheckpointingMixin.__init__(
            self, self.feature_extractor_fn, batchnorm_momentum, checkpoint_segments_size, use_activation_checkpointing
        )


class Resnet50_NoPreproc(ResNetCheckpointingMixin, ImageNetEncoder):
    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        use_activation_checkpointing: bool = False,
        checkpoint_segments_size: int = 2,
        batchnorm_momentum: Optional[float] = None,
    ) -> None:
        """
        :param tile_size: The size of the input tiles (default=224).
        :param n_channels: The number of channels in the input tiles (default=3).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        :param batchnorm_momentum: An optional momentum value to use for batch norm layers statistics updates when
            `use_activation_checkpointing` is True. If None (default), sqrt of the default momentum retrieved from the
            model is used to avoid running statistics from going out of sync due to activations checkpointing.
        """
        ImageNetEncoder.__init__(
            self,
            feature_extraction_model=resnet50,
            tile_size=tile_size,
            n_channels=n_channels,
            apply_imagenet_preprocessing=False,
            use_activation_checkpointing=use_activation_checkpointing,
        )
        ResNetCheckpointingMixin.__init__(
            self, self.feature_extractor_fn, batchnorm_momentum, checkpoint_segments_size, use_activation_checkpointing
        )


class SwinTransformerCheckpointingMixin:
    """Mixin class for checkpointing activations in SwinTransformer-based encoders."""

    def __init__(
        self,
        feature_extractor_fn: SwinTransformer,
        checkpoint_segments_size: int = 2,
    ) -> None:
        """
        :param feature_extractor_fn: A SwinTransformer model.
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        """
        assert isinstance(feature_extractor_fn, SwinTransformer), "Expected SwinTransformer model"
        self.feature_extractor_fn = feature_extractor_fn
        self.checkpoint_segments_size = checkpoint_segments_size

    def custom_patch_embedding_forward(self, images: torch.Tensor) -> torch.Tensor:
        """Custom patch partitioning checkpointing"""
        _, _, H, W = images.shape
        img_size = self.feature_extractor_fn.patch_embed.img_size
        assert (
            H == img_size[0] and W == img_size[1]
        ), f"Input image size ({H}*{W}) doesn't match model ({img_size[0]}*{img_size[1]})."
        images = checkpoint(self.feature_extractor_fn.patch_embed.proj, images)
        images = images.flatten(2).transpose(1, 2)  # BCHW -> BNC
        images = checkpoint(self.feature_extractor_fn.patch_embed.norm, images)
        return images

    def custom_forward(self, images: torch.Tensor) -> torch.Tensor:
        """Custom forward pass that uses activation checkpointing to save memory."""
        images = self.custom_patch_embedding_forward(images)
        # do not checkpoint dropout
        images = self.feature_extractor_fn.pos_drop(images)
        # sequential layers checkpointing
        assert isinstance(self.feature_extractor_fn.layers, nn.Sequential), "Expected nn.Sequential for layers."
        images = checkpoint_sequential(self.feature_extractor_fn.layers, self.checkpoint_segments_size, images)
        images = checkpoint(self.feature_extractor_fn.norm, images)
        # AvgPool the output of the last stage to get the feature maps
        images = images.mean(dim=1)
        return images


class SwinTransformer_NoPreproc(SwinTransformerCheckpointingMixin, ImageNetEncoder):
    """Swin Transformer encoder pretrained on ImageNet. This uses the `swin_tiny_patch4_window7_224` model which is a
    tiny version of the Swin Transformer model with a patch size of 4, a window size of 7 and input image size 224."""

    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        use_activation_checkpointing: bool = False,
        checkpoint_segments_size: int = 2,
    ) -> None:
        """
        :param tile_size: The size of the input tiles (default=224).
        :param n_channels: The number of channels in the input tiles (default=3).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        """
        ImageNetEncoder.__init__(
            self,
            feature_extraction_model=swin_tiny_patch4_window7_224,
            tile_size=tile_size,
            n_channels=n_channels,
            apply_imagenet_preprocessing=False,
            use_activation_checkpointing=use_activation_checkpointing,
        )
        SwinTransformerCheckpointingMixin.__init__(self, self.feature_extractor_fn, checkpoint_segments_size)

    def _get_encoder(self) -> Tuple[torch.nn.Module, int]:
        pretrained_model = self.create_feature_extractor_fn(pretrained=True, num_classes=0)
        return pretrained_model, pretrained_model.num_features  # type: ignore


class DenseNetCheckpointingMixin(BaseCheckpointingMixin):
    """Mixin class for checkpointing activations in DenseNet-based encoders."""

    def validate(self) -> None:
        """Validate that the feature extractor is a DenseNet model."""
        assert isinstance(self.feature_extractor_fn, DenseNet), "Expected DenseNet for feature_extractor_fn argument."

    def _set_batch_norm_momentum(self) -> None:
        """Set the momentum of batch norm layers in the DenseNet model to avoid running statistics from going out of
        sync due to activations checkpointing. The forward pass is applied twice which results in double updates of
        these statistics. We can workaround that by using sqrt of default momentum retrieved from the
        feature_extractor_fn.
        """
        if self.batchnorm_momentum is None:
            self.batchnorm_momentum = math.sqrt(self.feature_extractor_fn.features.norm0.momentum)

        self._set_all_batch_norm_momentum_in_block(self.feature_extractor_fn.features)

    def custom_forward(self, images: torch.Tensor) -> torch.Tensor:
        """Custom forward pass that uses activation checkpointing to save memory."""
        segments = self.checkpoint_segments_size
        features = checkpoint_sequential(self.feature_extractor_fn.features, segments, images)
        out = nn.functional.relu(features)
        out = nn.functional.adaptive_avg_pool2d(out, (1, 1))
        out = torch.flatten(out, 1)
        out = self.feature_extractor_fn.classifier(out)
        return out


class DenseNet121_NoPreproc(DenseNetCheckpointingMixin, ImageNetEncoder):
    """DenseNet121 encoder without imagenet preprocessing."""

    def __init__(
        self,
        tile_size: int = 224,
        n_channels: int = 3,
        use_activation_checkpointing: bool = False,
        checkpoint_segments_size: int = 2,
        batchnorm_momentum: Optional[float] = None,
    ) -> None:
        """
        :param tile_size: The size of the input tiles (default=224).
        :param n_channels: The number of channels in the input tiles (default=3).
        :param use_activation_checkpointing: Whether to checkpoint activations during forward pass. This can be used
            to reduce the memory required to store gradients by checkpointing the activations (default=False).
        :param checkpoint_segments_size: The size of checkpointed segments in sequential layers (default=2).
        :param batchnorm_momentum: An optional momentum value to use for batch norm layers statistics updates when
            `use_activation_checkpointing` is True. If None (default), sqrt of the default momentum retrieved from the
            model is used to avoid running statistics from going out of sync due to activations checkpointing.
        """
        ImageNetEncoder.__init__(
            self,
            feature_extraction_model=densenet121,
            tile_size=tile_size,
            n_channels=n_channels,
            apply_imagenet_preprocessing=False,
            use_activation_checkpointing=use_activation_checkpointing,
        )
        DenseNetCheckpointingMixin.__init__(
            self, self.feature_extractor_fn, batchnorm_momentum, checkpoint_segments_size, use_activation_checkpointing
        )

    def _get_encoder(self) -> Tuple[torch.nn.Module, int]:
        pretrained_model = self.create_feature_extractor_fn(pretrained=True)
        num_features: int = pretrained_model.classifier.in_features  # type: ignore
        pretrained_model.classifier = nn.Identity()
        return pretrained_model, num_features


class ImageNetSimCLREncoder(TileEncoder):
    """SimCLR encoder pretrained on ImageNet"""

    WEIGHTS_URL = (
        "https://pl-bolts-weights.s3.us-east-2.amazonaws.com/simclr/bolts_simclr_imagenet/simclr_imagenet.ckpt"
    )
    EMBEDDING_DIM = 2048

    def _get_preprocessing(self) -> Callable:
        return get_imagenet_preprocessing()

    def _get_encoder(self) -> Tuple[SimCLR, int]:
        simclr = SimCLR.load_from_checkpoint(self.WEIGHTS_URL, strict=False)
        simclr.freeze()
        return simclr, self.EMBEDDING_DIM


class SSLEncoder(TileEncoder):
    """SSL encoder trained on Azure ML"""

    def __init__(self, pl_checkpoint_path: Path, tile_size: int, n_channels: int = 3) -> None:
        """
        :param pl_checkpoint_path: The path of the downloaded checkpoint file.
        :param tile_size: Tile width/height, in pixels.
        :param n_channels: Number of channels in the tile (default=3).
        """
        self.pl_checkpoint_path = pl_checkpoint_path
        super().__init__(tile_size=tile_size, n_channels=n_channels)

    def _get_encoder(self) -> Tuple[torch.nn.Module, int]:
        try:
            from SSL.lightning_modules.ssl_classifier_module import SSLClassifier
            from SSL.utils import create_ssl_image_classifier
            from SSL import encoders
        except (ImportError, ModuleNotFoundError):
            raise ValueError("SSL not found. This class can only be used by using hi-ml from the GitHub source")
        model: SSLClassifier = create_ssl_image_classifier(  # type: ignore
            num_classes=1, freeze_encoder=True, pl_checkpoint_path=str(self.pl_checkpoint_path)  # dummy value
        )
        assert isinstance(model.encoder, encoders.SSLEncoder)
        return setup_feature_extractor(model.encoder.cnn_model, self.input_dim)

    def forward(self, x: T) -> T:
        x = super().forward(x)
        return x[-1] if isinstance(x, list) else x


class HistoSSLEncoder(TileEncoder):
    """HistoSSL encoder pretrained on multiple histological datasets

    Reference:
    - Ciga, Xu, Martel (2021). Self supervised contrastive learning for digital histopathology.
    arXiv:2011.13971
    """

    WEIGHTS_URL = (
        "https://github.com/ozanciga/self-supervised-histopathology/releases/"
        "download/tenpercent/tenpercent_resnet18.ckpt"
    )

    def _get_encoder(self) -> Tuple[torch.nn.Module, int]:
        resnet18_model = resnet18(pretrained=False)
        histossl_encoder = load_weights_to_model(self.WEIGHTS_URL, resnet18_model)
        return setup_feature_extractor(histossl_encoder, self.input_dim)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import List

import pandas as pd
from azureml.core import Run
from azureml.core.run import _OfflineRun

HIML_ROOT = Path(__file__).parent.parent.parent.parent.parent.absolute()
health_ml_root = HIML_ROOT / "hi-ml" / "src"
health_azure_root = HIML_ROOT / "hi-ml-azure" / "src"
sys.path.insert(0, str(health_ml_root))
sys.path.insert(0, str(health_azure_root))

from health_ml.utils.common_utils import df_to_json  # noqa: E402
from health_azure import aggregate_hyperdrive_metrics  # NOQA: E402
from health_azure.utils import get_aml_run_from_run_id, get_metrics_for_run  # NOQA: E402


def print_metrics(metrics_list: List[str], metrics_df: pd.DataFrame) -> None:
    """
    Given a DataFrame of metric names and corresponding values, and a list of the names metrics we
    want to print, attempts to locate each metric in the table and prints some summary statistics for it.

    :param metrics_list: The list of metrics to print values for.
    :param metrics_df: A pandas DataFrame representing a table of metric names and corresponding values.
    """
    for metric in metrics_list:
        if metric in metrics_df.index.values:
            mean = metrics_df.loc[[metric]].mean(axis=1)[metric]
            std = metrics_df.loc[[metric]].std(axis=1)[metric]
            print(f"{metric}: {round(mean,4)} ± {round(std,4)}")
        else:
            print(f"Metric {metric} not found in the Hyperdrive run metrics for run id {run_id}.")


def get_metrics_from_run(run: Run, metrics_list: List[str]) -> pd.DataFrame:
    """
    Gets a list of metrics from an Azure ML Run and returns them as a Pandas DataFrame.
    If the Run has children, then the DataFrame will contain the aggregated metrics across
    each of these children.

    :param run: An AzureML Run (which could be a HyperDriveRun, or a plain Run with no children)
    :param metrics_list: A list of metrics names to include.
    :return: A Pandas DataFrame containing metric names and corresponding values.
    """
    print(f"Getting metrics for run {run.id}")
    if isinstance(run, _OfflineRun):
        raise ValueError("Can't get metrics for an OfflineRun")
    if len(list(run.get_children())) > 0:
        metrics_df = aggregate_hyperdrive_metrics(
            child_run_arg_name="crossval_index", run=run, keep_metrics=metrics_list
        )
    else:
        metrics_df = get_metrics_for_run(run=run, keep_metrics=metrics_list)
    print_metrics(metrics_list, metrics_df)
    return metrics_df


def upload_regression_metrics_file_to_run(metrics_df: pd.DataFrame, run: Run) -> None:
    """
    For a given metrics DataFrame, creates a temporary local csv file and uploads its to the provided
    Azure ML Run. Note that this assumes there is an 'outputs' folder in your Run.

    :param metrics_df: the DataFrame of metrics that should be stored in a csv file and uploaded to the run.
    :param run: The AML Run to upload the metrics file to.
    """
    regression_results_dir = Path('/tmp') / run.id
    regression_results_dir.mkdir(exist_ok=True)
    metrics_json_output = regression_results_dir / "metrics.json"

    df_to_json(metrics_df, metrics_json_output)
    print("Uploading metrics file to AML Run")
    run.upload_file("outputs/regression_metrics.json", str(metrics_json_output))
    metrics_json_output.unlink()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--run", type=str, default='', help="The run id to retrieve metrics from")
    parser.add_argument(
        "--metrics_list", type=str, help="A comma-separated list of metrics names to retrieve from the AML Run"
    )
    parser.add_argument(
        "--upload_metrics_file",
        type=bool,
        default=True,
        help="If True, saves a json file of the metrics dataframe and uploads this to the AML Run",
    )
    args = parser.parse_args(sys.argv[1:])
    run_id = args.run
    metrics_list = args.metrics_list.split(",")

    run = get_aml_run_from_run_id(run_id)
    print(f"Run: {run}. Run id: {run.id}")
    metrics_df = get_metrics_from_run(run, metrics_list)
    if args.upload_metrics_file:
        upload_regression_metrics_file_to_run(metrics_df, run)



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import torch
import heapq

from torch import Tensor
from typing import Optional, List, Tuple, Dict, Union, Any
from collections import ChainMap

from health_cpath.utils.naming import ResultsKey, SlideKey, TileKey


class TileNode:
    """Data structure class for tile nodes used by `SlideNode` to store top and bottom tiles by attention scores."""

    def __init__(
        self,
        data: Tensor,
        attn: float,
        id: Optional[int] = None,
        left: Optional[float] = None,
        top: Optional[float] = None,
    ) -> None:
        """
        :param data: A tensor of tile data of shape (channels, height, width), e.g (3, 224, 224).
        :param attn: The attention score assigned to the tile.
        :param id: An optional tile id, defaults to None
        :param left: An optional tile coordinate at the LEFT (e.g. x_coor), defaults to None
        :param top: An optional tile coordinate at the TOP (e.g. y_coor), defaults to None
        """
        self.data = data.cpu()
        self.attn = attn
        self.id = id
        self.left = left
        self.top = top


class SlideNode:
    """Data structure class for slide nodes used by `TopBottomTilesHandler` to store top and bottom slides by
    probability score."""

    def __init__(
        self, slide_id: str, gt_prob_score: float, pred_prob_score: float, true_label: int, pred_label: int
    ) -> None:
        """
        :param gt_prob_score: The probability score assigned to ground truth label of slide node. This scalar defines
            the order of the slide_node among the nodes in the max/min heap.
        :param pred_prob_score: The probability score assigned to predicted label of slide node.
        :param slide_id: The slide id in the data cohort.
        :param true_label: The ground truth label of the slide node.
        :param pred_label: The label predicted by the model.
        """
        self.slide_id = slide_id
        self.gt_prob_score = gt_prob_score
        self.pred_prob_score = pred_prob_score
        self.true_label = true_label
        self.pred_label = pred_label
        self.top_tiles: List[TileNode] = []
        self.bottom_tiles: List[TileNode] = []

    def __lt__(self, other: "SlideNode") -> bool:
        return self.gt_prob_score < other.gt_prob_score

    def update_selected_tiles(self, tiles: Tensor, attn_scores: Tensor, num_top_tiles: int) -> None:
        """Update top and bottom k tiles values from a set of tiles and their assigned attention scores.

        :param tiles: A stack of tiles with shape (n_tiles, channels, height, width).
        :param attn_scores: A tensor of attention scores assigned by the deepmil model to tiles of shape (n_tiles, 1).
        :param num_top_tiles: The number of tiles to select as top and bottom tiles.
        """
        num_top_tiles = min(num_top_tiles, len(attn_scores))

        _, top_k_indices = torch.topk(attn_scores, k=num_top_tiles, largest=True, sorted=True)
        self.top_tiles = [TileNode(data=tiles[i], attn=attn_scores[i].item()) for i in top_k_indices]

        _, bottom_k_indices = torch.topk(attn_scores, k=num_top_tiles, largest=False, sorted=True)
        self.bottom_tiles = [TileNode(data=tiles[i], attn=attn_scores[i].item()) for i in bottom_k_indices]

    def _shallow_copy(self) -> "SlideNode":
        """Returns a shallow copy of the current slide node contaning only the slide_id and its probability score."""
        return SlideNode(self.slide_id, self.gt_prob_score, self.pred_prob_score, self.true_label, self.pred_label)


SlideOrTileKey = Union[SlideKey, TileKey]
SlideDict = Dict[int, List[SlideNode]]
TileDict = Dict[str, List[TileNode]]


class TilesSelector:
    """Class that manages selecting top and bottom tiles on the fly during validation or test of DeepMIL models."""

    def __init__(self, n_classes: int, num_slides: int = 10, num_tiles: int = 12) -> None:
        """
        :param n_classes: Number of MIL classes (set `n_classes=1` for binary).
        :param num_slides: Number of top and bottom slides to select to define top and bottom tiles based of
            prediction scores. Defaults to 10.
        :param num_tiles: Number of tiles to select as top and bottom tiles based on attn scores. Defaults to 12.
        """
        if num_slides > 0 and num_tiles == 0:
            raise ValueError(
                "You should use `num_top_tiles>0` to be able to select top and bottom tiles for `num_top_slides>0`. "
                "You can set `num_top_slides=0` to disable top and bottom tiles plotting."
            )

        self.n_classes = n_classes if n_classes > 1 else 2
        self.num_slides = num_slides
        self.num_tiles = num_tiles
        self.top_slides_heaps: SlideDict = self._initialise_slide_heaps()
        self.bottom_slides_heaps: SlideDict = self._initialise_slide_heaps()

    def _initialise_slide_heaps(self) -> SlideDict:
        return {class_id: [] for class_id in range(self.n_classes)}

    def _clear_cached_slides_heaps(self) -> None:
        self.top_slides_heaps = self._initialise_slide_heaps()
        self.bottom_slides_heaps = self._initialise_slide_heaps()

    def _reset_slides_heaps(self, new_top_slides_heaps: SlideDict, new_bottom_slides_heaps: SlideDict) -> None:
        self.top_slides_heaps = new_top_slides_heaps
        self.bottom_slides_heaps = new_bottom_slides_heaps

    def _update_label_slides(
        self,
        class_slides_heap: List[SlideNode],
        tiles: Tensor,
        attn_scores: Tensor,
        slide_node: SlideNode,
    ) -> None:
        """Update the selected slides of a given class label on the fly by updating the content of class_slides_heap.
        First, we push a shallow slide_node into the slides_heaps[gt_label]. The order in slides_heaps[gt_label] is
        defined by the slide_node.gt_prob_score that is positive in top_slides_heaps nodes and negative in
        bottom_slides_heaps nodes.
        Second, we check if we exceeded self.num_top_slides to be selected.
            If so, we update the slides_node top and bottom tiles only if it has been kept in the heap.
            Else, we haven't reached the heaps max_capacity so we save the top and bottom tiles.

        :param class_slides_heap: The class_slides_heap dict to be updated. Either self.top_slides_heaps or
            self.bottom_slides_heaps.
        :param tiles: Tiles of a given whole slide retrieved from the current validation or test batch.
            (n_tiles, channels, height, width)
        :param attn_scores: The tiles attention scores to determine top and bottom tiles. (n_tiles, )
        :param slide_node: A shallow version of slide_node that contains only slide_id and additional metadata.
        """
        heapq.heappush(class_slides_heap, slide_node)
        if len(class_slides_heap) == self.num_slides + 1:
            old_slide_node = heapq.heappop(class_slides_heap)
            if old_slide_node.slide_id != slide_node.slide_id:
                slide_node.update_selected_tiles(tiles, attn_scores, self.num_tiles)
        else:
            slide_node.update_selected_tiles(tiles, attn_scores, self.num_tiles)

    def update_slides_selection(self, batch: Dict[SlideOrTileKey, Any], results: Dict[ResultsKey, Any]) -> None:
        """Updates top and bottom slides heaps on the fly during validation or test.

        :param batch: The current validation/test batch that contains the slides info, corresponding tiles and
            additional metadata.
        :param results: Batch results that contain attention scores, probability scores and the true labels.
        """
        if self.num_slides > 0:
            slide_ids = batch[SlideKey.SLIDE_ID]
            if isinstance(slide_ids[0], list):
                slide_ids = [slide_id[0] for slide_id in slide_ids]  # to account for repetitions in tiles pipeline
            batch_size = len(batch[SlideKey.IMAGE])
            for i in range(batch_size):
                gt_label = results[ResultsKey.TRUE_LABEL][i].item()
                pred_label = results[ResultsKey.PRED_LABEL][i].item()
                gt_prob_score = results[ResultsKey.CLASS_PROBS][:, gt_label][i].item()
                pred_prob_score = results[ResultsKey.CLASS_PROBS][:, pred_label][i].item()
                tiles = batch[SlideKey.IMAGE][i]
                attn_scores = results[ResultsKey.BAG_ATTN][i].squeeze(0)
                if pred_label == gt_label:
                    self._update_label_slides(
                        class_slides_heap=self.top_slides_heaps[gt_label],
                        tiles=tiles,
                        attn_scores=attn_scores,
                        slide_node=SlideNode(
                            slide_id=slide_ids[i],
                            gt_prob_score=gt_prob_score,
                            pred_prob_score=pred_prob_score,
                            true_label=gt_label,
                            pred_label=pred_label,
                        ),
                    )
                elif pred_label != gt_label:
                    self._update_label_slides(
                        class_slides_heap=self.bottom_slides_heaps[gt_label],
                        tiles=tiles,
                        attn_scores=attn_scores,
                        slide_node=SlideNode(
                            slide_id=slide_ids[i],
                            # negative score for bottom slides to reverse order in max heap
                            gt_prob_score=-gt_prob_score,
                            pred_prob_score=pred_prob_score,
                            true_label=gt_label,
                            pred_label=pred_label,
                        ),
                    )

    def _shallow_copy_slides_heaps(self, slides_heaps: SlideDict) -> SlideDict:
        """Returns a shallow copy of slides heaps to be synchronised across devices.

        :param slides_heaps: The slides_heaps dict to be shallow copied. Either self.top_slides_heaps or
            self.bottom_slides_heaps.
        """
        shallow_slides_heaps_copy: SlideDict = {}
        for class_id, slide_nodes in slides_heaps.items():
            shallow_slides_heaps_copy[class_id] = [slide_node._shallow_copy() for slide_node in slide_nodes]
        return shallow_slides_heaps_copy

    def _reduce_slides_heaps_list(self, world_size: int, slides_heaps_list: List[SlideDict]) -> SlideDict:
        """Reduces the list of slides_heaps gathered across devices into a single slides_heaps that contain top or
        bottom shallow slide nodes.

        :param world_size: The number of devices in the ddp context.
        :param slides_heaps_list: A list of slides heaps gathered across devices.
        :return: A reduced version of slides_heaps_list to a single slides_heaps containing only top or bottom slides.
        """
        slides_heaps: SlideDict = self._initialise_slide_heaps()
        for class_id in range(self.n_classes):
            for rank in range(world_size):
                for slide_node in slides_heaps_list[rank][class_id]:
                    heapq.heappush(slides_heaps[class_id], slide_node)
                    if len(slides_heaps[class_id]) == self.num_slides + 1:
                        heapq.heappop(slides_heaps[class_id])
        return slides_heaps

    @staticmethod
    def _gather_dictionaries(world_size: int, dicts: Dict, return_list: bool = False) -> Union[List[Dict], Dict]:
        """Gathers python dictionaries accross devices.

        :param world_size: The number of devices in the ddp context.
        :param dicts: The dictionaries to be gathered across devices.
        :param return_list: Flag to return the gathered dictionaries as a list, defaults to False
        :return: Either a list of dictionaries or a reduced version of this list as a single dictionary.
        """
        dicts_list = [None] * world_size  # type: ignore
        torch.distributed.all_gather_object(dicts_list, dicts)
        if return_list:
            return dicts_list  # type: ignore
        return dict(ChainMap(*dicts_list))  # type: ignore

    def _aggregate_shallow_slides_heaps(self, world_size: int, shallow_slides_heaps: SlideDict) -> SlideDict:
        """Gathers shallow slides heaps across devices and reduces the gathered slides_heaps_list into the new select
        slides across devices.

        :param world_size: The number of devices in the ddp context.
        :param shallow_slides_heaps: Reference to shallow slides heaps to be gathered across devices.
        :return: A reduced slides_heaps conatining only the best slide nodes across all devices.
        """
        slides_heaps_list: List[SlideDict] = self._gather_dictionaries(  # type: ignore
            world_size, shallow_slides_heaps, return_list=True
        )
        return self._reduce_slides_heaps_list(world_size=world_size, slides_heaps_list=slides_heaps_list)

    def _collect_tiles_for_selected_slides_on_device(
        self, new_slides_heaps: SlideDict, slides_heaps: SlideDict
    ) -> Tuple[TileDict, TileDict]:
        """Select top and bottom tiles from top or bottom slides_heaps in current device.

        :param new_slides_heaps: Selected top or bottom slides_heaps that contains shallow slide nodes.
        :param slides_heaps: The slides_heaps dict from where to select tiles. Either self.top_slides_heaps or
            self.bottom_slides_heaps.
        :return: Two dictionaries containing slide_ids and their corresponding top and bottom tiles.
        """
        top_tiles = {}
        bottom_tiles = {}
        for class_id in range(self.n_classes):
            for new_slide_node in new_slides_heaps[class_id]:
                for slide_node in slides_heaps[class_id]:
                    if slide_node.slide_id == new_slide_node.slide_id:
                        top_tiles[new_slide_node.slide_id] = slide_node.top_tiles
                        bottom_tiles[new_slide_node.slide_id] = slide_node.bottom_tiles
                        break
        return top_tiles, bottom_tiles

    def _update_shallow_slides_heaps_with_top_bottom_tiles(
        self,
        slides_heaps: SlideDict,
        top_tiles: TileDict,
        bottom_tiles: TileDict,
    ) -> None:
        """Update shallow version of slides heaps with top and bottom tiles gathered across devices.

        :param slides_heaps: The slides_heaps to update. Either self.top_slides_heaps or self.bottom_slides_heaps.
        :param top_tiles: Top tiles dictionary with slide_ids as keys.
        :param bottom_tiles: Bottom tiles dictionary with slide_ids as keys.
        """
        for class_id in range(self.n_classes):
            for slide_node in slides_heaps[class_id]:
                slide_node.top_tiles = top_tiles[slide_node.slide_id]
                slide_node.bottom_tiles = bottom_tiles[slide_node.slide_id]

    def gather_selected_tiles_across_devices(self) -> None:
        """Gathers top and bottom tiles across devices in ddp context. For each of top and bottom slides heaps:
        1- make a shallow copy of top and bottom slides_heaps
        2- gather best shallow slides across gpus
        3- select top and bottom tiles available in each device
        4- gather these tiles across devices
        5- update the synchronized shallow slide nodes across devices with their top and bottom tiles
        """
        if torch.distributed.is_initialized() and self.num_slides > 0:
            world_size = torch.distributed.get_world_size()
            if world_size > 1:
                shallow_top_slides_heaps = self._shallow_copy_slides_heaps(slides_heaps=self.top_slides_heaps)
                shallow_bottom_slides_heaps = self._shallow_copy_slides_heaps(slides_heaps=self.bottom_slides_heaps)

                agg_top_slides_heaps = self._aggregate_shallow_slides_heaps(world_size, shallow_top_slides_heaps)
                agg_bottom_slides_heaps = self._aggregate_shallow_slides_heaps(world_size, shallow_bottom_slides_heaps)

                top_slides_top_tiles, top_slides_bottom_tiles = self._collect_tiles_for_selected_slides_on_device(
                    new_slides_heaps=agg_top_slides_heaps, slides_heaps=self.top_slides_heaps
                )
                bot_slides_top_tiles, bot_slides_bottom_tiles = self._collect_tiles_for_selected_slides_on_device(
                    new_slides_heaps=agg_bottom_slides_heaps, slides_heaps=self.bottom_slides_heaps
                )

                self._reset_slides_heaps(
                    new_top_slides_heaps=agg_top_slides_heaps, new_bottom_slides_heaps=agg_bottom_slides_heaps
                )

                top_tiles: TileDict = self._gather_dictionaries(world_size, top_slides_top_tiles)  # type: ignore
                bottom_tiles: TileDict = self._gather_dictionaries(world_size, top_slides_bottom_tiles)  # type: ignore
                self._update_shallow_slides_heaps_with_top_bottom_tiles(self.top_slides_heaps, top_tiles, bottom_tiles)

                top_tiles: TileDict = self._gather_dictionaries(world_size, bot_slides_top_tiles)  # type: ignore
                bottom_tiles: TileDict = self._gather_dictionaries(world_size, bot_slides_bottom_tiles)  # type: ignore
                self._update_shallow_slides_heaps_with_top_bottom_tiles(
                    self.bottom_slides_heaps, top_tiles, bottom_tiles
                )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import pandas as pd


def extract_fields(row: pd.Series) -> dict:
    # Paths are structured as follows:
    # "CRC_DX_[TEST|TRAIN]/[MSS|MSIMUT]/blk-{tile_id}-{slide_id}-01Z-00-DX1.png"
    # - tile_id is an uppercase string of 12 letters
    # - slide_id is "TCGA-XX-XXXX"
    parts = row.image.split('/')
    return {
        'slide_id': parts[2][17:29],
        'tile_id': parts[2][4:16],
        'image': row.image,
        'label': {'MSS': 0, 'MSIMUT': 1}[parts[1]],
        'split': parts[0][7:].lower(),
    }
